// Author: Jianxin Wu (wujx2001@gmail.com)

#include <algorithm>
#include <numeric>
#include <vector>
#include <string>
#include <iostream>
#include <fstream>
#include <climits>

#include "mdarray.h"
#include "util.h"
#include "IntImage.h"
#include "HistogramKernel.h"
#include "LinearKernel.h"
#include "Codebook.h"
#include "Features.h"

using namespace std;

CodeBook::CodeBook(const DESCRIPTOR_TYPE _feature_type,const bool _useSobel,const int _resizeWidth,const int _windowSize,const int _L1_norm)
    :verbose(false),feature_type(_feature_type),useSobel(_useSobel),validcenters(0),resizeWidth(_resizeWidth),windowSize(_windowSize),L1_norm(_L1_norm)
{
    assert(ValidFeatureType(feature_type));
    assert(resizeWidth>=0 && windowSize>0 && L1_norm>0);
}

CodeBook::~CodeBook()
{
}

BaseFeature* CodeBook::FeatureEngine(const DESCRIPTOR_TYPE feature_type,const bool useSobel,const int L1_norm)
{
    assert(ValidFeatureType(feature_type) && L1_norm>0);
    BaseFeature* feature = NULL;
    switch(feature_type)
    {
        case DESCRIPTOR_CENTRIST:
            feature = new CENTRIST(useSobel,L1_norm);
            break;
        case DESCRIPTOR_SIFT:
            feature = new SiftFeature(useSobel,L1_norm);
            break;
        case DESCRIPTOR_LOCAL:
            feature = new LocalFeature(useSobel);
            break;
        case DESCRIPTOR_SYM_CENTRIST:
            feature = new SymCENTRIST(useSobel,L1_norm);
            break;
        default:
            std::cout<<"Feature type is wrong."<<std::endl;
            break;
    }
    return feature;
}

void CodeBook::TranslateOneImage(const char* filename,const int stepSize,const int splitlevel,
                                    double* p,const int fsize,const bool normalize,const double ratio,int scales) const
{
    // we could put 'feature' as a member of the CodeBook class. However, put it here makes parallel processing (OpenMP) possible
    BaseFeature* feature = FeatureEngine(feature_type,useSobel,L1_norm); // generate the right feature extractor

    const double scale_change = 0.75; // ratio to shrink the image to a smaller scale & weights at different scaled version of the image
    double scale = 1; // weight of codewords generated in different resized version of the image

    const IntImage<double>* img = &feature->AssignFile(filename,resizeWidth); // assign an image to the feature extractor
    std::fill(p,p+fsize,0.0);
    while(scales>=0)
    {
        int match;
        for(int i=1;i+windowSize<img->nrow;i+=stepSize)
        {
            for(int j=1;j+windowSize<img->ncol;j+=stepSize)
            {
                match = Find_Nearest(i,i+windowSize,j,j+windowSize,feature); // find the right codeword
                if(match<0) continue; // NOTE:special case -- this image patch is nearly uniform
                p[match]+=scale; // depth 0 of the hierarchy
                int x,y;
                if(splitlevel>0)
                {
                    x = (i+windowSize/2)*4/img->nrow;
                    y = (j+windowSize/2)*4/img->ncol;
                    if(x>0 && x<3 && y>0 && y<3) p[validcenters+match] += scale; // depth 1 of the hierarchy
                    x /= 2;
                    y /= 2;
                    p[validcenters*(2+x*2+y)+match] += scale; // depth 1 of the hierarchy
                }
                if(splitlevel>1)
                {
                    x = (i+windowSize/2)*8/img->nrow;
                    y = (j+windowSize/2)*8/img->ncol;
                    if(x>0 && x<7 && y>0 && y<7) p[validcenters*(6+(x-1)/2*3+(y-1)/2)+match] += scale; // depth 2 of the hierarchy
                    x /= 2;
                    y /= 2;
                    p[validcenters*(15+x*4+y)+match] += scale; // depth 2 of the hierarchy
                }
            }
        }
        img = &feature->ResizeImage(scale_change); // shrink the image if number of 'scales' is not reached
        scales--;
        scale *= (1.0/scale_change/scale_change); // set weight of codewords generated in different resized images
    }
    delete feature; feature = NULL;

    if(normalize) Normalize_L1(p,validcenters);
    if(splitlevel>=1) // normalize if necessary, and weight the codewords differently depending on depth in the hierarchy
    {
        if(normalize) for(int j=0;j<5;j++) Normalize_L1(p+validcenters+validcenters*j,validcenters);
        for(int j=0;j<5*validcenters;j++) p[validcenters+j] *= ratio;
    }
    if(splitlevel>=2)
    {
        if(normalize) for(int j=0;j<25;j++) Normalize_L1(p+validcenters+validcenters*5+validcenters*j,validcenters);
        for(int j=0;j<25*validcenters;j++) p[validcenters+validcenters*5+j] *= (ratio*ratio);
    }
}

IplImage* CodeBook::TranslateOneImage(const char* filename,const int K) const
{
    CvScalar* colors = new CvScalar[K];
    my_srand(123456);
    for(int i=0;i<256;i++) colors[i] = cvScalar(my_rand()%K,my_rand()%K,my_rand()%K,my_rand()%K);

    BaseFeature* feature = FeatureEngine(feature_type,useSobel,L1_norm);
    const IntImage<double>& img = feature->AssignFile(filename,0);
    IplImage* result = cvCreateImage(cvSize(img.ncol,img.nrow),IPL_DEPTH_8U,3);
    const CvScalar black = cvScalar(0,0,0,0);
    for(int i=0;i<img.nrow;i++) for(int j=0;j<img.ncol;j++) cvSet2D(result,i,j,black);
    for(int i=1;i+windowSize<img.nrow;i++)
    {
        for(int j=1;j+windowSize<img.ncol;j++)
        {
            int match = Find_Nearest(i,i+windowSize,j,j+windowSize,feature);
            assert(match<K);
            cvSet2D(result,i+windowSize/2,j+windowSize/2,colors[match]);
        }
    }
    delete feature; feature = NULL;
	delete[] colors;
    return result;
}

LinearCodes::LinearCodes(const DESCRIPTOR_TYPE _feature_type,const bool _useSobel,const int _resizeWidth,const int _windowSize,const int _L1_norm,const bool _useMedian)
    :CodeBook(_feature_type,_useSobel,_resizeWidth,_windowSize,_L1_norm),useMedian(_useMedian)
{
}

LinearCodes::~LinearCodes()
{
}

void LinearCodes::GenerateClusterData(const std::vector<const char*>& files,const int stepSize)
{
    BaseFeature* feature = FeatureEngine(feature_type,useSobel,L1_norm);
    features.Create(1000,feature->Length());
    int added = 0;
    for(unsigned int imgindex=0;imgindex<files.size();imgindex++)
    {
        const IntImage<double>& img = feature->AssignFile(files[imgindex],resizeWidth); // assign image to the feature extractor
        int num_subwin = ((img.nrow-2)/stepSize+1) * ((img.ncol-2)/stepSize+1); // number of features generated for this image
        if(added+num_subwin>features.nrow) // increase capacity of 'features' if necessary
            features.AdjustCapacity(max(features.nrow*3/2,added+num_subwin));
        for(int i=1;i+windowSize<img.nrow;i+=stepSize)
        {
            for(int j=1;j+windowSize<img.ncol;j+=stepSize)
            {   // LinearCodes use real valued ('D'--double) version of features
                const double* p = feature->D_feature(i,i+windowSize,j,j+windowSize);
                if(p[0]<0) continue; // NOTE: special case -- image patch nearly uniform
                std::copy(p,p+feature->Length(),features.p[added]);
                added++;
            }
        }
    }
    features.AdjustCapacity(added);
    delete feature; feature = NULL;
}

// set data for training purpose
void LinearCodes::SetData(Array2d<double>& data)
{
    features = data;
}

void LinearCodes::SetData(Array2d<int>& data)
{
    features.Create(data.nrow,data.ncol);
    for(int i=0;i<data.nrow;i++) for(int j=0;j<data.ncol;j++) features.p[i][j] = data.p[i][j];
}

int LinearCodes::GenerateCodeWords(const int K,const bool oneclassSVM,const int maxiteration)
{
    Array2dC<int> labels(1,features.nrow);
    Array2dC<int> counts(1,K);
    rho.Create(1,K);
    Linear_Kernel_Kmeans kmeans(features,K,0.0001,maxiteration,false,labels.buf,counts.buf,eval,rho.buf,useMedian,verbose);
    validcenters = kmeans.KernelKmeans();
    if(oneclassSVM) validcenters = kmeans.OneClassSVM();
    features.Clear();
    return validcenters;
}

int LinearCodes::Find_Nearest(const int x1,const int x2,const int y1,const int y2,BaseFeature* feature) const
{
    const double* p = feature->D_feature(x1,x2,y1,y2);
    if(p[0]<0) return -1; // NOTE: special case
    else
    return Linear_Find_Nearest(p,eval,feature->Length(),validcenters,rho.buf,useMedian);
}

HistogramCodes::HistogramCodes(const DESCRIPTOR_TYPE _feature_type,const bool _useSobel,const int _resizeWidth,const int _windowSize,const int _L1_norm,const int _upper_bound)
    :CodeBook(_feature_type,_useSobel,_resizeWidth,_windowSize,_L1_norm),upper_bound(_upper_bound)
{
}

HistogramCodes::~HistogramCodes()
{
}

void HistogramCodes::GenerateClusterData(const std::vector<const char*>& files,const int stepSize)
{
    BaseFeature* feature = FeatureEngine(feature_type,useSobel,L1_norm);
    features.Create(1000,feature->Length());
    int added = 0;
    for(unsigned int imgindex=0;imgindex<files.size();imgindex++)
    {
        const IntImage<double>& img = feature->AssignFile(files[imgindex],resizeWidth);
        int num_subwin = ((img.nrow-2)/stepSize+1) * ((img.ncol-2)/stepSize+1);
        if(added+num_subwin>features.nrow) features.AdjustCapacity(max(features.nrow*3/2,added+num_subwin));
        for(int i=1;i+windowSize<img.nrow;i+=stepSize)
        {
            for(int j=1;j+windowSize<img.ncol;j+=stepSize)
            {   // LinearCodes use discrete valued ('I'--integer) version of features
                const int* p = feature->I_feature(i,i+windowSize,j,j+windowSize);
                if(p[0]<0) continue; // NOTE: special case
                std::copy(p,p+feature->Length(),features.p[added]);
                added++;
            }
        }
    }
    features.AdjustCapacity(added);
    delete feature; feature = NULL;
}

void HistogramCodes::SetData(Array2d<int>& data)
// set data for training purpose
{
    features = data;
}

int HistogramCodes::GenerateCodeWords(const int K,const bool oneclassSVM,const int maxiteration)
{
    Array2dC<int> labels(1,features.nrow);
    Array2dC<int> counts(1,K);
    rho.Create(1,K);
    Histogram_Kernel_Kmeans kmeans(features,K,upper_bound,0.0001,maxiteration,false,labels.buf,counts.buf,eval,rho.buf,verbose);
    validcenters = kmeans.KernelKmeans();
    if(oneclassSVM) validcenters = kmeans.OneClassSVM();
    features.Clear();
    return validcenters;
}

int HistogramCodes::Find_Nearest(const int x1,const int x2,const int y1,const int y2,BaseFeature* feature) const
{
    const int* f = feature->I_feature(x1,x2,y1,y2);
    if(f[0]<0) return -1; // NOTE: special case
    else
    return Histogram_Find_Nearest(f,eval,feature->Length(),validcenters,upper_bound,rho.buf);
}

LocalCodes::LocalCodes(const bool _useSobel,const int _resizeWidth)
    :CodeBook(DESCRIPTOR_CENTRIST,_useSobel,_resizeWidth,3,INT_MAX)
{
    validcenters = 256;
}

LocalCodes::~LocalCodes()
{
}

int LocalCodes::Find_Nearest(const int x1,const int x2,const int y1,const int y2,BaseFeature* feature) const
{
    // we assuem 3x3 patches, so x2 & y2 are not useful
    return CensusTransform<double>(feature->Image().p,x1+1,y1+1);
}
